@@ -1,5 +1,5 @@
 /*
-    Copyright 2005-2012 Intel Corporation.  All Rights Reserved.
+    Copyright 2005-2014 Intel Corporation.  All Rights Reserved.
 
     This file is part of Threading Building Blocks.
 
@@ -45,7 +45,7 @@
 
 #include <iterator>
 #include <utility>      // Need std::pair
-#include <functional>
+#include <functional>   // Need std::equal_to (in ../concurrent_unordered_*.h)
 #include <string>       // For tbb_hasher
 #include <cstring>      // Need std::memset
 
@@ -56,6 +56,11 @@
 #include "../atomic.h"
 #include "../tbb_exception.h"
 #include "../tbb_allocator.h"
+#include "tbb/atomic.h"
+
+#if __TBB_INITIALIZER_LISTS_PRESENT
+    #include <initializer_list>
+#endif
 
 namespace tbb {
 namespace interface5 {
@@ -189,6 +194,7 @@ bool operator!=( const solist_iterator<Solist,T>& i, const solist_iterator<Solis
 // Forward type and class definitions
 typedef size_t sokey_t;
 
+
 // Forward list in which elements are sorted in a split-order
 template <typename T, typename Allocator>
 class split_ordered_list
@@ -215,6 +221,10 @@ public:
     // Node that holds the element in a split-ordered list
     struct node : tbb::internal::no_assign
     {
+    private:
+        // for compilers that try to generate default constructors though they are not needed.
+        node();  // VS 2008, 2010, 2012
+    public:
         // Initialize the node with the given order key
         void init(sokey_t order_key) {
             my_order_key = order_key;
@@ -230,7 +240,7 @@ public:
         nodeptr_t atomic_set_next(nodeptr_t new_node, nodeptr_t current_node)
         {
             // Try to change the next pointer on the current element to a new element, only if it still points to the cached next
-            nodeptr_t exchange_node = (nodeptr_t) __TBB_CompareAndSwapW((void *) &my_next, (uintptr_t)new_node, (uintptr_t)current_node);
+            nodeptr_t exchange_node = tbb::internal::as_atomic(my_next).compare_and_swap(new_node, current_node);
 
             if (exchange_node == current_node) // TODO: why this branch?
             {
@@ -274,15 +284,7 @@ public:
     // Allocate a new node with the given order key; used to allocate dummy nodes
     nodeptr_t create_node(sokey_t order_key) {
         nodeptr_t pnode = my_node_allocator.allocate(1);
-
-        __TBB_TRY {
-            new(static_cast<void*>(&pnode->my_element)) T();
-            pnode->init(order_key);
-        } __TBB_CATCH(...) {
-            my_node_allocator.deallocate(pnode, 1);
-            __TBB_RETHROW();
-        }
-
+        pnode->init(order_key);
         return (pnode);
     }
 
@@ -383,8 +385,8 @@ public:
             return;
         }
 
-        std::swap(my_element_count, other.my_element_count);
-        std::swap(my_head, other.my_head);
+            std::swap(my_element_count, other.my_element_count);
+            std::swap(my_head, other.my_head);
     }
 
     // Split-order list functions
@@ -412,7 +414,7 @@ public:
     }
 
     static sokey_t get_safe_order_key(const raw_const_iterator& it) {
-        if( !it.get_node_ptr() ) return sokey_t(~0U);
+        if( !it.get_node_ptr() )  return ~sokey_t(0);
         return it.get_node_ptr()->get_order_key();
     }
 
@@ -464,7 +466,7 @@ public:
 
     // Erase an element using the allocator
     void destroy_node(nodeptr_t pnode) {
-        my_node_allocator.destroy(pnode);
+        if (!pnode->is_dummy()) my_node_allocator.destroy(pnode);
         my_node_allocator.deallocate(pnode, 1);
     }
 
@@ -721,6 +723,17 @@ protected:
         return (*this);
     }
 
+#if __TBB_INITIALIZER_LISTS_PRESENT
+    //! assignment operator from initializer_list
+    concurrent_unordered_base& operator=(std::initializer_list<value_type> const& il)
+    {
+        this->clear();
+        this->insert(il.begin(),il.end());
+        return (*this);
+    }
+#endif //# __TBB_INITIALIZER_LISTS_PRESENT
+
+
     ~concurrent_unordered_base() {
         // Delete all node segments
         internal_clear();
@@ -821,11 +834,16 @@ public:
                 sokey_t end_key = solist_t::get_safe_order_key(my_end_node);
                 size_t mid_bucket = __TBB_ReverseBits( begin_key + (end_key-begin_key)/2 ) % my_table.my_number_of_buckets;
                 while ( !my_table.is_initialized(mid_bucket) ) mid_bucket = my_table.get_parent(mid_bucket);
-                my_midpoint_node = my_table.my_solist.first_real_iterator(my_table.get_bucket( mid_bucket ));
-                if( my_midpoint_node == my_begin_node )
-                    my_midpoint_node = my_end_node;
-#if TBB_USE_ASSERT
+                if(__TBB_ReverseBits(mid_bucket) > begin_key) {
+                    // found a dummy_node between begin and end
+                    my_midpoint_node = my_table.my_solist.first_real_iterator(my_table.get_bucket( mid_bucket ));
+                }
                 else {
+                    // didn't find a dummy node between begin and end.
+                    my_midpoint_node = my_end_node;
+                }
+#if TBB_USE_ASSERT
+                {
                     sokey_t mid_key = solist_t::get_safe_order_key(my_midpoint_node);
                     __TBB_ASSERT( begin_key < mid_key, "my_begin_node is after my_midpoint_node" );
                     __TBB_ASSERT( mid_key <= end_key, "my_midpoint_node is after my_end_node" );
@@ -1122,7 +1140,7 @@ private:
         if (!is_initialized(bucket))
             init_bucket(bucket);
 
-        size_type new_count;
+        size_type new_count = 0;
         order_key = split_order_key_regular(order_key);
         raw_iterator it = get_bucket(bucket);
         raw_iterator last = my_solist.raw_end();
@@ -1295,8 +1313,8 @@ private:
         // Grow the table by a factor of 2 if possible and needed
         if ( ((float) total_elements / (float) current_size) > my_maximum_bucket_size )
         {
-            // Double the size of the hash only if size has not changed inbetween loads
-            __TBB_CompareAndSwapW((uintptr_t*)&my_number_of_buckets, uintptr_t(2u*current_size), uintptr_t(current_size) );
+            // Double the size of the hash only if size has not changed in between loads
+            my_number_of_buckets.compare_and_swap(2u*current_size, current_size);
             //Simple "my_number_of_buckets.compare_and_swap( current_size<<1, current_size );" does not work for VC8
             //due to overzealous compiler warnings in /Wp64 mode
         }
@@ -1342,7 +1360,7 @@ private:
             raw_iterator * new_segment = my_allocator.allocate(sz);
             std::memset(new_segment, 0, sz*sizeof(raw_iterator));
 
-            if (__TBB_CompareAndSwapW((void *) &my_buckets[segment], (uintptr_t)new_segment, 0) != 0)
+            if (my_buckets[segment].compare_and_swap( new_segment, NULL) != NULL)
                 my_allocator.deallocate(new_segment, sz);
         }
 
@@ -1369,7 +1387,7 @@ private:
 
     // A dummy order key has its original hash value reversed and the last bit unset
     sokey_t split_order_key_dummy(sokey_t order_key) const {
-        return __TBB_ReverseBits(order_key) & ~(0x1);
+        return __TBB_ReverseBits(order_key) & ~sokey_t(0x1);
     }
 
     // Shared variables
@@ -1384,7 +1402,7 @@ private:
 #endif
 
 //! Hash multiplier
-static const size_t hash_multiplier = tbb::internal::size_t_select(2654435769U, 11400714819323198485ULL);
+static const size_t hash_multiplier = tbb::internal::select_size_t_constant<2654435769U, 11400714819323198485ULL>::value;
 } // namespace internal
 //! @endcond
 //! Hasher functions
