@@ -1,5 +1,5 @@
 /*
-    Copyright 2005-2012 Intel Corporation.  All Rights Reserved.
+    Copyright 2005-2014 Intel Corporation.  All Rights Reserved.
 
     This file is part of Threading Building Blocks.
 
@@ -59,13 +59,45 @@ bool freeRawMemory (void *object, size_t size) {
     return UnmapMemory(object, size);
 }
 
-void ExtMemoryPool::reportHugePageStatus(bool available)
+void HugePagesStatus::registerAllocation(bool gotPage)
 {
+    if (gotPage) {
+        if (!wasObserved)
+            FencedStore(wasObserved, 1);
+    } else
+        FencedStore(enabled, 0);
     // reports huge page status only once
-    if (FencedLoad(needHugePageStatus)
-        && AtomicCompareExchange(needHugePageStatus, 0, 1))
-        fprintf(stderr, "TBBmalloc: scalable allocator\t%s\n",
-                available? "use huge pages" : "no huge pages" );
+    if (needActualStatusPrint
+        && AtomicCompareExchange(needActualStatusPrint, 0, 1))
+        doPrintStatus(gotPage, "available");
+}
+
+void HugePagesStatus::registerReleasing(size_t size)
+{
+    // We: 1) got huge page at least once,
+    // 2) something that looks like a huge page is been released,
+    // and 3) user requested huge pages,
+    // so a huge page might be available at next allocation.
+    // TODO: keep page status in regions and use exact check here
+    // Use isPowerOfTwoMultiple because it's faster then generic reminder.
+    if (FencedLoad(wasObserved) && isPowerOfTwoMultiple(size, pageSize))
+        FencedStore(enabled, requestedMode.get());
+}
+
+void HugePagesStatus::printStatus() {
+    doPrintStatus(requestedMode.get(), "requested");
+    if (requestedMode.get()) { // report actual status iff requested
+        if (pageSize)
+            FencedStore(needActualStatusPrint, 1);
+        else
+            doPrintStatus(/*state=*/false, "available");
+    }
+}
+
+void HugePagesStatus::doPrintStatus(bool state, const char *stateName)
+{
+    fprintf(stderr, "TBBmalloc: huge pages\t%s%s\n",
+            state? "" : "not ", stateName);
 }
 
 void *Backend::getRawMem(size_t &size) const
@@ -76,21 +108,20 @@ void *Backend::getRawMem(size_t &size) const
     }
     // try to get them at 1st allocation and still use, if successful
     // if 1st try is unsuccessful, no more trying
-    if (ExtMemoryPool::useHugePages) {
-        size_t hugeSize = alignUpGeneric(size, ExtMemoryPool::hugePageSize);
-        if (void *res = getRawMemory(hugeSize, /*hugePages=*/true)) {
+    if (FencedLoad(hugePages.enabled)) {
+        size_t hugeSize = alignUpGeneric(size, hugePages.getSize());
+        void *res = getRawMemory(hugeSize, /*hugePages=*/true);
+        hugePages.registerAllocation(res);
+        if (res) {
             size = hugeSize;
-            if (!ExtMemoryPool::seenHugePages)
-                ExtMemoryPool::seenHugePages = true;
-            ExtMemoryPool::reportHugePageStatus(/*available=*/true);
+            AtomicAdd((intptr_t&)totalMemSize, size);
             return res;
         }
-        ExtMemoryPool::useHugePages = false;
     }
-    ExtMemoryPool::reportHugePageStatus(/*available=*/false);
     size_t granSize = alignUpGeneric(size, extMemPool->granularity);
     if (void *res = getRawMemory(granSize, /*hugePages=*/false)) {
         size = granSize;
+        AtomicAdd((intptr_t&)totalMemSize, size);
         return res;
     }
     return NULL;
@@ -98,16 +129,11 @@ void *Backend::getRawMem(size_t &size) const
 
 void Backend::freeRawMem(void *object, size_t size) const
 {
+    AtomicAdd((intptr_t&)totalMemSize, -size);
     if (extMemPool->userPool())
         (*extMemPool->rawFree)(extMemPool->poolId, object, size);
     else {
-        // We got huge page at least once and
-        // something that looks like a huge page is releasing,
-        // so a huge page might be available at next allocation.
-        // TODO: keep page status in regions and use exact check here
-        if (ExtMemoryPool::seenHugePages && !(size % ExtMemoryPool::hugePageSize)
-            && !ExtMemoryPool::useHugePages)
-            ExtMemoryPool::useHugePages = true;
+        hugePages.registerReleasing(size);
         freeRawMemory(object, size);
     }
 }
@@ -281,8 +307,8 @@ FreeBlock *CoalRequestQ::getAll()
 // If the remaining free space would stay in the same bin,
 //     split the block without removing it.
 // If the free space should go to other bin(s), remove the block.
-// alignedBin is true, if all blocks in the bin has slab-aligned right side.
-FreeBlock *Backend::IndexedBins::getBlock(int binIdx, BackendSync *sync,
+// alignedBin is true, if all blocks in the bin have slab-aligned right side.
+FreeBlock *Backend::IndexedBins::getFromBin(int binIdx, BackendSync *sync,
                 size_t size, bool needAlignedRes, bool alignedBin, bool wait,
                 int *binLocked)
 {
@@ -301,6 +327,8 @@ try_next:
         for (FreeBlock *curr = b->head; curr; curr = curr->next) {
             size_t szBlock = curr->tryLockBlock();
             if (!szBlock) {
+                // block is locked, re-do bin lock, as there is no place to spin
+                // while block coalescing
                 goto try_next;
             }
 
@@ -327,7 +355,7 @@ try_next:
             if (fBlock) {
                 // consume must be called before result of removing from a bin
                 // is visible externally.
-                sync->consume();
+                sync->blockConsumed();
                 if (alignedBin && needAlignedRes &&
                     Backend::sizeToBin(szBlock-size) == Backend::sizeToBin(szBlock)) {
                     // free remainder of fBlock stay in same bin,
@@ -358,8 +386,37 @@ try_next:
     return fBlock;
 }
 
+bool Backend::IndexedBins::tryReleaseRegions(int binIdx, Backend *backend)
+{
+    Bin *b = &freeBins[binIdx];
+    FreeBlock *fBlockList = NULL;
+
+    // got all blocks from the bin and re-do coalesce on them
+    // to release single-block regions
+try_next:
+    if (b->head) {
+        MallocMutex::scoped_lock binLock(b->tLock);
+        for (FreeBlock *curr = b->head; curr; ) {
+            size_t szBlock = curr->tryLockBlock();
+            if (!szBlock)
+                goto try_next;
+
+            FreeBlock *next = curr->next;
+
+            b->removeBlock(curr);
+            curr->sizeTmp = szBlock;
+            curr->nextToFree = fBlockList;
+            fBlockList = curr;
+            curr = next;
+        }
+    }
+    return backend->coalescAndPutList(fBlockList, /*forceCoalescQDrop=*/true);
+}
+
 void Backend::Bin::removeBlock(FreeBlock *fBlock)
 {
+    MALLOC_ASSERT(fBlock->next||fBlock->prev||fBlock==head,
+                  "Detected that a block is not in the bin.");
     if (head == fBlock)
         head = fBlock->next;
     if (tail == fBlock)
@@ -370,21 +427,30 @@ void Backend::Bin::removeBlock(FreeBlock *fBlock)
         fBlock->next->prev = fBlock->prev;
 }
 
-void Backend::IndexedBins::addBlock(int binIdx, FreeBlock *fBlock, size_t blockSz)
+void Backend::IndexedBins::addBlock(int binIdx, FreeBlock *fBlock, size_t blockSz, bool addToTail)
 {
     Bin *b = &freeBins[binIdx];
 
     fBlock->myBin = binIdx;
     fBlock->aligned = toAlignedBin(fBlock, blockSz);
-    fBlock->prev = NULL;
+    fBlock->next = fBlock->prev = NULL;
     {
         MallocMutex::scoped_lock scopedLock(b->tLock);
-        fBlock->next = b->head;
-        b->head = fBlock;
-        if (fBlock->next)
-            fBlock->next->prev = fBlock;
-        if (!b->tail)
+        if (addToTail) {
+            fBlock->prev = b->tail;
             b->tail = fBlock;
+            if (fBlock->prev)
+                fBlock->prev->next = fBlock;
+            if (!b->head)
+                b->head = fBlock;
+        } else {
+            fBlock->next = b->head;
+            b->head = fBlock;
+            if (fBlock->next)
+                fBlock->next->prev = fBlock;
+            if (!b->tail)
+                b->tail = fBlock;
+        }
     }
     bitMask.set(binIdx, true);
 }
@@ -447,250 +513,256 @@ bool ExtMemoryPool::regionsAreReleaseable() const
     return !keepAllMemory && !delayRegsReleasing;
 }
 
-// try to allocate num blocks of size Bytes from particular "generic" bin
-// needAlignedRes is true if result must be slab-aligned
-FreeBlock *Backend::getFromBin(int binIdx, int num, size_t size, bool needAlignedRes,
-                               int *binLocked)
+FreeBlock *Backend::splitUnalignedBlock(FreeBlock *fBlock, int num, size_t size,
+                                        bool needAlignedBlock)
 {
-    FreeBlock *fBlock =
-        freeLargeBins.getBlock(binIdx, &bkndSync, num*size, needAlignedRes,
-                               /*alignedBin=*/false, /*wait=*/false, binLocked);
-    if (fBlock) {
-        if (needAlignedRes) {
-            size_t fBlockSz = fBlock->sizeTmp;
-            uintptr_t fBlockEnd = (uintptr_t)fBlock + fBlockSz;
-            FreeBlock *newB = alignUp(fBlock, slabSize);
-            FreeBlock *rightPart = (FreeBlock*)((uintptr_t)newB + num*size);
+    const size_t totalSize = num*size;
+    if (needAlignedBlock) {
+        size_t fBlockSz = fBlock->sizeTmp;
+        uintptr_t fBlockEnd = (uintptr_t)fBlock + fBlockSz;
+        FreeBlock *newB = alignUp(fBlock, slabSize);
+        FreeBlock *rightPart = (FreeBlock*)((uintptr_t)newB + totalSize);
 
-            // Space to use is in the middle,
-            // ... return free right part
-            if ((uintptr_t)rightPart != fBlockEnd) {
-                rightPart->initHeader();  // to prevent coalescing rightPart with fBlock
-                coalescAndPut(rightPart, fBlockEnd - (uintptr_t)rightPart);
-            }
-            // ... and free left part
-            if (newB != fBlock) {
-                newB->initHeader(); // to prevent coalescing fBlock with newB
-                coalescAndPut(fBlock, (uintptr_t)newB - (uintptr_t)fBlock);
-            }
+        // Space to use is in the middle,
+        // ... return free right part
+        if ((uintptr_t)rightPart != fBlockEnd) {
+            rightPart->initHeader();  // to prevent coalescing rightPart with fBlock
+            coalescAndPut(rightPart, fBlockEnd - (uintptr_t)rightPart);
+        }
+        // ... and free left part
+        if (newB != fBlock) {
+            newB->initHeader(); // to prevent coalescing fBlock with newB
+            coalescAndPut(fBlock, (uintptr_t)newB - (uintptr_t)fBlock);
+        }
 
-            fBlock = newB;
-            MALLOC_ASSERT(isAligned(fBlock, slabSize), ASSERT_TEXT);
+        fBlock = newB;
+        MALLOC_ASSERT(isAligned(fBlock, slabSize), ASSERT_TEXT);
+    } else {
+        if (size_t splitSz = fBlock->sizeTmp - totalSize) {
+            // split block and return free right part
+            FreeBlock *splitB = (FreeBlock*)((uintptr_t)fBlock + totalSize);
+            splitB->initHeader();
+            coalescAndPut(splitB, splitSz);
+        }
+    }
+    bkndSync.blockReleased();
+    FreeBlock::markBlocks(fBlock, num, size);
+    return fBlock;
+}
+
+FreeBlock *Backend::splitAlignedBlock(FreeBlock *fBlock, int num, size_t size,
+                                      bool needAlignedBlock)
+{
+    if (fBlock->sizeTmp != num*size) { // i.e., need to split the block
+        FreeBlock *newAlgnd;
+        size_t newSz;
+
+        if (needAlignedBlock) {
+            newAlgnd = fBlock;
+            fBlock = (FreeBlock*)((uintptr_t)newAlgnd + newAlgnd->sizeTmp
+                                  - num*size);
+            MALLOC_ASSERT(isAligned(fBlock, slabSize), "Invalid free block");
+            fBlock->initHeader();
+            newSz = newAlgnd->sizeTmp - num*size;
         } else {
-            if (size_t splitSz = fBlock->sizeTmp - num*size) {
-                // split block and return free right part
-                FreeBlock *splitB = (FreeBlock*)((uintptr_t)fBlock + num*size);
-                splitB->initHeader();
-                coalescAndPut(splitB, splitSz);
-            }
+            newAlgnd = (FreeBlock*)((uintptr_t)fBlock + num*size);
+            newSz = fBlock->sizeTmp - num*size;
+            newAlgnd->initHeader();
         }
-        bkndSync.signal();
-        FreeBlock::markBlocks(fBlock, num, size);
+        coalescAndPut(newAlgnd, newSz);
     }
-
+    bkndSync.blockReleased();
+    MALLOC_ASSERT(!needAlignedBlock || isAligned(fBlock, slabSize),
+                  "Expect to get aligned block, if one was requested.");
+    FreeBlock::markBlocks(fBlock, num, size);
     return fBlock;
 }
 
-// try to allocate size Byte block from any of slab-aligned spaces.
-// needAlignedRes is true if result must be slab-aligned
-FreeBlock *Backend::getFromAlignedSpace(int binIdx, int num, size_t size,
-                                        bool needAlignedRes, bool wait, int *binLocked)
+inline size_t Backend::getMaxBinnedSize() const
 {
-    FreeBlock *fBlock =
-        freeAlignedBins.getBlock(binIdx, &bkndSync, num*size, needAlignedRes,
-                                 /*alignedBin=*/true, wait, binLocked);
-
-    if (fBlock) {
-        if (fBlock->sizeTmp != num*size) { // i.e., need to split the block
-            FreeBlock *newAlgnd;
-            size_t newSz;
-
-            if (needAlignedRes) {
-                newAlgnd = fBlock;
-                fBlock = (FreeBlock*)((uintptr_t)newAlgnd + newAlgnd->sizeTmp
-                                      - num*size);
-                MALLOC_ASSERT(isAligned(fBlock, slabSize), "Invalid free block");
-                fBlock->initHeader();
-                newSz = newAlgnd->sizeTmp - num*size;
-            } else {
-                newAlgnd = (FreeBlock*)((uintptr_t)fBlock + num*size);
-                newSz = fBlock->sizeTmp - num*size;
-                newAlgnd->initHeader();
-            }
-            coalescAndPut(newAlgnd, newSz);
-        }
-        bkndSync.signal();
-        MALLOC_ASSERT(!needAlignedRes || isAligned(fBlock, slabSize), ASSERT_TEXT);
-        FreeBlock::markBlocks(fBlock, num, size);
-    }
-    return fBlock;
-}
-
-void Backend::correctMaxRequestSize(size_t requestSize)
-{
-    // There were huge pages, but they are all consumed;
-    // switch back to maxBinned_SmallPage then.
-    size_t maxBinned = getMaxBinnedSize();
-    for (size_t oldMaxReq = maxRequestedSize; oldMaxReq > maxBinned; ) {
-        oldMaxReq = AtomicCompareExchange((intptr_t&)maxRequestedSize,
-                                           maxBinned, oldMaxReq);
-        maxBinned = getMaxBinnedSize();
-    }
-
-    // Find maximal requested size limited by getMaxBinnedSize()
-    if (requestSize < getMaxBinnedSize()) {
-        for (size_t oldMaxReq = maxRequestedSize;
-             requestSize > oldMaxReq && requestSize < getMaxBinnedSize(); ) {
-            size_t val = AtomicCompareExchange((intptr_t&)maxRequestedSize,
-                                               requestSize, oldMaxReq);
-            if (val == oldMaxReq)
-                break;
-            oldMaxReq = val;
-        }
-    }
-}
-
-inline size_t Backend::getMaxBinnedSize()
-{
-    return ExtMemoryPool::useHugePages && !inUserPool()?
+    return hugePages.wasObserved && !inUserPool()?
         maxBinned_HugePage : maxBinned_SmallPage;
 }
 
-bool Backend::askMemFromOS(size_t blockSize, intptr_t startModifiedCnt,
-                           int *lockedBinsThreshold,
-                           int numOfLockedBins, bool *largeBinsUpdated)
+inline bool Backend::MaxRequestComparator::operator()(size_t oldMaxReq,
+                                                      size_t requestSize) const
 {
-    size_t maxBinSize = 0;
+    return requestSize > oldMaxReq && requestSize < backend->getMaxBinnedSize();
+}
+
+FreeBlock *Backend::askMemFromOS(size_t blockSize, intptr_t startModifiedCnt,
+                                 int *lockedBinsThreshold, int numOfLockedBins)
+{
+    FreeBlock *block = (FreeBlock*)VALID_BLOCK_IN_BIN;
 
     // Another thread is modifying backend while we can't get the block.
     // Wait while it leaves and re-do the scan
     // before trying other ways to extend the backend.
-    if (bkndSync.waitTillSignalled(startModifiedCnt)
+    if (bkndSync.waitTillBlockReleased(startModifiedCnt)
         // semaphore is protecting adding more more memory from OS
         || memExtendingSema.wait())
-        return true;
+        return (FreeBlock*)VALID_BLOCK_IN_BIN;
 
     if (startModifiedCnt != bkndSync.getNumOfMods()) {
         memExtendingSema.signal();
-        return true;
+        return (FreeBlock*)VALID_BLOCK_IN_BIN;
     }
     // To keep objects below maxBinnedSize, region must be larger then that.
     // So trying to balance between too small regions (that leads to
     // fragmentation) and too large ones (that leads to excessive address
     // space consumption). If region is "quite large", allocate only one,
-    // to prevent fragmentation. It supposely doesn't hurt perfromance,
+    // to prevent fragmentation. It supposedly doesn't hurt performance,
     // because the object requested by user is large.
     const size_t maxBinned = getMaxBinnedSize();
     const size_t regSz_sizeBased = blockSize>=maxBinned?
         blockSize : alignUp(4*maxRequestedSize, 1024*1024);
     if (blockSize == slabSize || blockSize == numOfSlabAllocOnMiss*slabSize
         || regSz_sizeBased < maxBinned) {
-        for (unsigned idx=0; idx<4; idx++) {
-            size_t binSize = addNewRegion(maxBinned, /*exact=*/false);
-            if (!binSize)
-                break;
-            if (binSize > maxBinSize)
-                maxBinSize = binSize;
-        }
+        // For this size of blocks, add NUM_OF_REG "advance" regions in bin,
+        // and return one as a result.
+        // TODO: add to bin first, because other threads can use them right away.
+        // This must be done carefully, because blocks in bins can be released
+        // in releaseCachesToLimit().
+        const unsigned NUM_OF_REG = 3;
+        block = addNewRegion(maxBinned, /*exact=*/false, /*addToBin=*/false);
+        if (block)
+            for (unsigned idx=0; idx<NUM_OF_REG; idx++)
+                if (! addNewRegion(maxBinned, /*exact=*/false, /*addToBin=*/true))
+                    break;
     } else {
-        *largeBinsUpdated = true;
-        maxBinSize = addNewRegion(regSz_sizeBased, /*exact=*/true);
+        // if huge pages enabled and blockSize > maxBinned, rest of space up to
+        // huge page alignment is unusable, because single user object sits
+        // in an region.
+        block = addNewRegion(regSz_sizeBased, /*exact=*/true, /*addToBin=*/false);
     }
     memExtendingSema.signal();
-    askMemFromOSCounter.OSasked();
 
-    // When blockSize >= maxBinnedSize, and getRawMem failed
-    // for this allocation, allocation in bins
-    // is our last chance to fulfil the request.
-    // Sadly, size is larger then max bin, so have to give up.
-    if (maxBinSize && maxBinSize < blockSize)
-        return false;
-
-    if (!maxBinSize) { // no regions have been added, try to clean cache
-        if (extMemPool->hardCachesCleanup())
-            *largeBinsUpdated = true;
-        else {
-            if (bkndSync.waitTillSignalled(startModifiedCnt))
-                return true;
-            // OS can't give us more memory, but we have some in locked bins
-            if (*lockedBinsThreshold && numOfLockedBins) {
-                *lockedBinsThreshold = 0;
-                return true;
-            }
-            return false;
+    // no regions found, try to clean cache
+    if (!block || block == (FreeBlock*)VALID_BLOCK_IN_BIN) {
+            // something released from caches
+        if (extMemPool->hardCachesCleanup()
+            // ..or can use blocks that are in processing now
+            || bkndSync.waitTillBlockReleased(startModifiedCnt))
+            return (FreeBlock*)VALID_BLOCK_IN_BIN;
+        // OS can't give us more memory, but we have some in locked bins
+        if (*lockedBinsThreshold && numOfLockedBins) {
+            *lockedBinsThreshold = 0;
+            return (FreeBlock*)VALID_BLOCK_IN_BIN;
         }
+        return NULL; // nothing found, give up
     }
-    return true;
+    // after asking memory from OS, release caches if we above the memory limits
+    releaseCachesToLimit();
+
+    return block;
+}
+
+void Backend::releaseCachesToLimit()
+{
+    if (!memSoftLimit || totalMemSize <= memSoftLimit)
+        return;
+    size_t locTotalMemSize, locMemSoftLimit;
+
+    scanCoalescQ(/*forceCoalescQDrop=*/false);
+    if (extMemPool->softCachesCleanup() &&
+        (locTotalMemSize = FencedLoad((intptr_t&)totalMemSize)) <=
+        (locMemSoftLimit = FencedLoad((intptr_t&)memSoftLimit)))
+        return;
+    // clean global large-object cache, if this is not enough, clean local caches
+    // do this in several tries, because backend fragmentation can prevent
+    // region from releasing
+    for (int cleanLocal = 0; cleanLocal<2; cleanLocal++)
+        while (cleanLocal?
+               extMemPool->allLocalCaches.cleanup(extMemPool, /*cleanOnlyUnused=*/true)
+               : extMemPool->loc.decreasingCleanup())
+            if ((locTotalMemSize = FencedLoad((intptr_t&)totalMemSize)) <=
+                (locMemSoftLimit = FencedLoad((intptr_t&)memSoftLimit)))
+                return;
+    // last chance to match memSoftLimit
+    extMemPool->hardCachesCleanup();
+}
+
+FreeBlock *Backend::IndexedBins::
+    findBlock(int nativeBin, BackendSync *sync, size_t size,
+              bool resSlabAligned, bool alignedBin, int *numOfLockedBins)
+{
+    for (int i=getMinNonemptyBin(nativeBin); i<freeBinsNum; i=getMinNonemptyBin(i+1))
+        if (FreeBlock *block = getFromBin(i, sync, size, resSlabAligned, alignedBin,
+                                          /*wait=*/false, numOfLockedBins))
+            return block;
+
+    return NULL;
 }
 
 // try to allocate size Byte block in available bins
 // needAlignedRes is true if result must be slab-aligned
-FreeBlock *Backend::genericGetBlock(int num, size_t size, bool needAlignedRes)
+FreeBlock *Backend::genericGetBlock(int num, size_t size, bool needAlignedBlock)
 {
-    // after (soft|hard)CachesCleanup we can get memory in large bins,
-    // while after addNewRegion only in slab-aligned bins. This flag
-    // is for large bins update status.
-    bool largeBinsUpdated = true;
     FreeBlock *block = NULL;
     const size_t totalReqSize = num*size;
+    // no splitting after requesting new region, asks exact size
     const int nativeBin = sizeToBin(totalReqSize);
     // If we found 2 or less locked bins, it's time to ask more memory from OS.
     // But nothing can be asked from fixed pool. And we prefer wait, not ask
     // for more memory, if block is quite large.
     int lockedBinsThreshold = extMemPool->fixedPool || size>=maxBinned_SmallPage? 0 : 2;
 
-    correctMaxRequestSize(totalReqSize);
+    // Find maximal requested size limited by getMaxBinnedSize()
+    AtomicUpdate(maxRequestedSize, totalReqSize, MaxRequestComparator(this));
     scanCoalescQ(/*forceCoalescQDrop=*/false);
 
     for (;;) {
         const intptr_t startModifiedCnt = bkndSync.getNumOfMods();
         int numOfLockedBins;
 
-        for (;;) {
+        do {
             numOfLockedBins = 0;
 
             // TODO: try different bin search order
-            if (needAlignedRes) {
+            if (needAlignedBlock) {
+                block = freeAlignedBins.findBlock(nativeBin, &bkndSync, num*size,
+                                    /*needAlignedBlock=*/true, /*alignedBin=*/true,
+                                    &numOfLockedBins);
                 if (!block)
-                    for ( int i=freeAlignedBins.getMinNonemptyBin(nativeBin);
-                          i<freeBinsNum; i=freeAlignedBins.getMinNonemptyBin(i+1) ){
-                        block = getFromAlignedSpace(i, num, size, /*needAlignedRes=*/true, /*wait=*/false, &numOfLockedBins);
-                        if (block) break;
-                    }
-                if (!block && largeBinsUpdated)
-                    for ( int i=freeLargeBins.getMinNonemptyBin(nativeBin);
-                          i<freeBinsNum; i=freeLargeBins.getMinNonemptyBin(i+1) ){
-                        block = getFromBin(i, num, size, /*needAlignedRes=*/true, &numOfLockedBins);
-                        if (block) break;
-                    }
+                    block = freeLargeBins.findBlock(nativeBin, &bkndSync, num*size,
+                                    /*needAlignedBlock=*/true, /*alignedBin=*/false,
+                                    &numOfLockedBins);
             } else {
-                if (!block && largeBinsUpdated)
-                    for ( int i=freeLargeBins.getMinNonemptyBin(nativeBin);
-                          i<freeBinsNum; i=freeLargeBins.getMinNonemptyBin(i+1) ){
-                        block = getFromBin(i, num, size, /*needAlignedRes=*/false, &numOfLockedBins);
-                        if (block) break;
-                    }
+                block = freeLargeBins.findBlock(nativeBin, &bkndSync, num*size,
+                                    /*needAlignedBlock=*/false, /*alignedBin=*/false,
+                                    &numOfLockedBins);
                 if (!block)
-                    for ( int i=freeAlignedBins.getMinNonemptyBin(nativeBin);
-                          i<freeBinsNum; i=freeAlignedBins.getMinNonemptyBin(i+1) ){
-                        block = getFromAlignedSpace(i, num, size, /*needAlignedRes=*/false, /*wait=*/false, &numOfLockedBins);
-                        if (block) break;
-                    }
+                    block = freeAlignedBins.findBlock(nativeBin, &bkndSync, num*size,
+                                    /*needAlignedBlock=*/false, /*alignedBin=*/true,
+                                    &numOfLockedBins);
             }
-            if (block || numOfLockedBins<=lockedBinsThreshold)
-                break;
-        }
+        } while (!block && numOfLockedBins>lockedBinsThreshold);
+
         if (block)
             break;
 
-        largeBinsUpdated = scanCoalescQ(/*forceCoalescQDrop=*/true);
-        largeBinsUpdated = extMemPool->softCachesCleanup() || largeBinsUpdated;
-        if (!largeBinsUpdated) {
-            if (!askMemFromOS(totalReqSize, startModifiedCnt, &lockedBinsThreshold,
-                              numOfLockedBins, &largeBinsUpdated))
+        if (!(scanCoalescQ(/*forceCoalescQDrop=*/true)
+              | extMemPool->softCachesCleanup())) {
+            // bins are not updated,
+            // only remaining possibility is to ask for more memory
+            block =
+                askMemFromOS(totalReqSize, startModifiedCnt, &lockedBinsThreshold,
+                             numOfLockedBins);
+            if (!block)
                 return NULL;
+            if (block != (FreeBlock*)VALID_BLOCK_IN_BIN) {
+                // size can be increased in askMemFromOS, that's why >=
+                MALLOC_ASSERT(block->sizeTmp >= size, ASSERT_TEXT);
+                break;
+            }
+            // valid block somewhere in bins, let's find it
+            block = NULL;
         }
     }
+    if (block)
+        block = toAlignedBin(block, block->sizeTmp)?
+            splitAlignedBlock(block, num, size, needAlignedBlock) :
+            splitUnalignedBlock(block, num, size, needAlignedBlock);
     return block;
 }
 
@@ -729,18 +801,19 @@ void Backend::putBackRefSpace(void *b, size_t size, bool rawMemUsed)
 
 void Backend::removeBlockFromBin(FreeBlock *fBlock)
 {
-    if (fBlock->myBin != Backend::NO_BIN)
+    if (fBlock->myBin != Backend::NO_BIN) {
         if (fBlock->aligned)
             freeAlignedBins.lockRemoveBlock(fBlock->myBin, fBlock);
         else
             freeLargeBins.lockRemoveBlock(fBlock->myBin, fBlock);
+    }
 }
 
 void Backend::genericPutBlock(FreeBlock *fBlock, size_t blockSz)
 {
-    bkndSync.consume();
+    bkndSync.blockConsumed();
     coalescAndPut(fBlock, blockSz);
-    bkndSync.signal();
+    bkndSync.blockReleased();
 }
 
 void AllLargeBlocksList::add(LargeMemoryBlock *lmb)
@@ -764,20 +837,6 @@ void AllLargeBlocksList::remove(LargeMemoryBlock *lmb)
         lmb->gPrev->gNext = lmb->gNext;
 }
 
-void AllLargeBlocksList::removeAll(Backend *backend)
-{
-    LargeMemoryBlock *next, *lmb = loHead;
-    loHead = NULL;
-
-    for (; lmb; lmb = next) {
-        next = lmb->gNext;
-        // nothing left to AllLargeBlocksList::remove
-        lmb->gNext = lmb->gPrev = NULL;
-        removeBackRef(lmb->backRefIdx);
-        backend->putLargeBlock(lmb);
-    }
-}
-
 void Backend::putLargeBlock(LargeMemoryBlock *lmb)
 {
     if (extMemPool->mustBeAddedToGlobalLargeBlockList())
@@ -785,6 +844,13 @@ void Backend::putLargeBlock(LargeMemoryBlock *lmb)
     genericPutBlock((FreeBlock *)lmb, lmb->unalignedSize);
 }
 
+void Backend::returnLargeObject(LargeMemoryBlock *lmb)
+{
+    removeBackRef(lmb->backRefIdx);
+    putLargeBlock(lmb);
+    STAT_increment(getThreadId(), ThreadCommonCounters, freeLargeObj);
+}
+
 void Backend::releaseRegion(MemRegion *memRegion)
 {
     {
@@ -886,8 +952,9 @@ FreeBlock *Backend::doCoalesc(FreeBlock *fBlock, MemRegion **mRegion)
     return resBlock;
 }
 
-void Backend::coalescAndPutList(FreeBlock *list, bool forceCoalescQDrop, bool doStat)
+bool Backend::coalescAndPutList(FreeBlock *list, bool forceCoalescQDrop)
 {
+    bool regionReleased = false;
     FreeBlock *helper;
     MemRegion *memRegion;
 
@@ -905,6 +972,7 @@ void Backend::coalescAndPutList(FreeBlock *list, bool forceCoalescQDrop, bool do
                 if (toRet->blockInBin)
                     removeBlockFromBin(toRet);
                 releaseRegion(memRegion);
+                regionReleased = true;
                 continue;
             } else // add block from empty region to end of bin,
                 addToTail = true; // preserving for exact fit
@@ -933,9 +1001,10 @@ void Backend::coalescAndPutList(FreeBlock *list, bool forceCoalescQDrop, bool do
             // It's not a leak because the block later can be coalesced.
             if (currSz >= minBinnedSize) {
                 toRet->sizeTmp = currSz;
-                if (!(toAligned?
-                      freeAlignedBins.tryAddBlock(bin, toRet, addToTail) :
-                      freeLargeBins.tryAddBlock(bin, toRet, addToTail))) {
+                IndexedBins *target = toAligned? &freeAlignedBins : &freeLargeBins;
+                if (forceCoalescQDrop) {
+                    target->addBlock(bin, toRet, toRet->sizeTmp, addToTail);
+                } else if (!target->tryAddBlock(bin, toRet, addToTail)) {
                     coalescQ.putBlock(toRet);
                     continue;
                 }
@@ -950,6 +1019,7 @@ void Backend::coalescAndPutList(FreeBlock *list, bool forceCoalescQDrop, bool do
         toRet->setMeFree(currSz);
         toRet->rightNeig(currSz)->setLeftFree(currSz);
     }
+    return regionReleased;
 }
 
 // Coalesce fBlock and add it back to a bin;
@@ -959,7 +1029,7 @@ void Backend::coalescAndPut(FreeBlock *fBlock, size_t blockSz)
     fBlock->sizeTmp = blockSz;
     fBlock->nextToFree = NULL;
 
-    coalescAndPutList(fBlock, /*forceCoalescQDrop=*/false, /*doStat=*/false);
+    coalescAndPutList(fBlock, /*forceCoalescQDrop=*/false);
 }
 
 bool Backend::scanCoalescQ(bool forceCoalescQDrop)
@@ -967,11 +1037,11 @@ bool Backend::scanCoalescQ(bool forceCoalescQDrop)
     FreeBlock *currCoalescList = coalescQ.getAll();
 
     if (currCoalescList)
-        coalescAndPutList(currCoalescList, forceCoalescQDrop, /*doStat=*/true);
+        coalescAndPutList(currCoalescList, forceCoalescQDrop);
     return currCoalescList;
 }
 
-FreeBlock *Backend::findBlockInRegion(MemRegion *region)
+FreeBlock *Backend::findBlockInRegion(MemRegion *region, size_t exactBlockSize)
 {
     FreeBlock *fBlock;
     size_t blockSz;
@@ -981,7 +1051,8 @@ FreeBlock *Backend::findBlockInRegion(MemRegion *region)
     if (region->exact) {
         fBlock = (FreeBlock *)alignUp((uintptr_t)region + sizeof(MemRegion),
                                       largeObjectAlignment);
-        fBlockEnd = lastFreeBlock;
+        fBlockEnd = (uintptr_t)fBlock + exactBlockSize;
+        MALLOC_ASSERT(fBlockEnd <= lastFreeBlock, ASSERT_TEXT);
     } else { // right bound is slab-aligned, keep LastFreeBlock after it
         fBlock = (FreeBlock *)((uintptr_t)region + sizeof(MemRegion));
         fBlockEnd = alignDown(lastFreeBlock, slabSize);
@@ -1001,7 +1072,7 @@ FreeBlock *Backend::findBlockInRegion(MemRegion *region)
 
 // startUseBlock adds free block to a bin, the block can be used and
 // even released after this, so the region must be added to regionList already
-void Backend::startUseBlock(MemRegion *region, FreeBlock *fBlock)
+void Backend::startUseBlock(MemRegion *region, FreeBlock *fBlock, bool addToBin)
 {
     size_t blockSz = region->blockSz;
     fBlock->initHeader();
@@ -1014,15 +1085,23 @@ void Backend::startUseBlock(MemRegion *region, FreeBlock *fBlock)
     lastBl->myBin = NO_BIN;
     lastBl->memRegion = region;
 
-    unsigned targetBin = sizeToBin(blockSz);
-    if (!region->exact && toAlignedBin(fBlock, blockSz)) {
-        freeAlignedBins.addBlock(targetBin, fBlock, blockSz);
+    if (addToBin) {
+        unsigned targetBin = sizeToBin(blockSz);
+        // during adding advance regions, register bin for a largest block in region
+        advRegBins.registerBin(targetBin);
+        if (!region->exact && toAlignedBin(fBlock, blockSz)) {
+            freeAlignedBins.addBlock(targetBin, fBlock, blockSz, /*addToTail=*/false);
+        } else {
+            freeLargeBins.addBlock(targetBin, fBlock, blockSz, /*addToTail=*/false);
+        }
     } else {
-        freeLargeBins.addBlock(targetBin, fBlock, blockSz);
+        // to match with blockReleased() in split(Unaligned|Aligned)Block
+        bkndSync.blockConsumed();
+        fBlock->sizeTmp = fBlock->tryLockBlock();
     }
 }
 
-size_t Backend::addNewRegion(size_t rawSize, bool exact)
+FreeBlock *Backend::addNewRegion(size_t size, bool exact, bool addToBin)
 {
     // to guarantee that header is not overwritten in used blocks
     MALLOC_ASSERT(sizeof(BlockMutexes) <= sizeof(BlockI), ASSERT_TEXT);
@@ -1032,25 +1111,26 @@ size_t Backend::addNewRegion(size_t rawSize, bool exact)
     // "exact" means that not less than rawSize for block inside the region.
     // Reserve space for region header, worst case alignment
     // and last block mark.
-    if (exact)
-        rawSize += sizeof(MemRegion) + largeObjectAlignment 
-                +  FreeBlock::minBlockSize + sizeof(LastFreeBlock);
+    size_t rawSize = exact?
+        size + sizeof(MemRegion) + largeObjectAlignment
+             +  FreeBlock::minBlockSize + sizeof(LastFreeBlock)
+        : size;
 
     MemRegion *region = (MemRegion*)getRawMem(rawSize);
     if (!region) return 0;
     if (rawSize < sizeof(MemRegion)) {
         if (!extMemPool->fixedPool)
             freeRawMem(region, rawSize);
-        return 0;
+        return NULL;
     }
 
     region->exact = exact;
     region->allocSz = rawSize;
-    FreeBlock *fBlock = findBlockInRegion(region);
+    FreeBlock *fBlock = findBlockInRegion(region, size);
     if (!fBlock) {
         if (!extMemPool->fixedPool)
             freeRawMem(region, rawSize);
-        return 0;
+        return NULL;
     }
     // adding to global list of all regions
     {
@@ -1061,12 +1141,15 @@ size_t Backend::addNewRegion(size_t rawSize, bool exact)
         if (regionList->next)
             regionList->next->prev = regionList;
     }
-    // copy it here, as just after starting to use region it might be released
-    size_t blockSz = region->blockSz;
+    startUseBlock(region, fBlock, addToBin);
+    bkndSync.binsModified();
+    return addToBin? (FreeBlock*)VALID_BLOCK_IN_BIN : fBlock;
+}
 
-    startUseBlock(region, fBlock);
-    bkndSync.pureSignal();
-    return blockSz;
+bool Backend::bootstrap(ExtMemoryPool *extMemoryPool)
+{
+    extMemPool = extMemoryPool;
+    return addNewRegion(2*1024*1024, /*exact=*/false, /*addToBin=*/true);
 }
 
 void Backend::reset()
@@ -1079,11 +1162,12 @@ void Backend::reset()
 
     freeLargeBins.reset();
     freeAlignedBins.reset();
+    advRegBins.reset();
 
     for (curr = regionList; curr; curr = curr->next) {
-        FreeBlock *fBlock = findBlockInRegion(curr);
+        FreeBlock *fBlock = findBlockInRegion(curr, curr->blockSz);
         MALLOC_ASSERT(fBlock, "A memory region unexpectedly got smaller");
-        startUseBlock(curr, fBlock);
+        startUseBlock(curr, fBlock, /*addToBin=*/true);
     }
 }
 
@@ -1091,6 +1175,10 @@ bool Backend::destroy()
 {
     // no active threads are allowed in backend while destroy() called
     verify();
+
+    freeLargeBins.reset();
+    freeAlignedBins.reset();
+
     while (regionList) {
         MemRegion *helper = regionList->next;
         if (inUserPool())
@@ -1104,6 +1192,24 @@ bool Backend::destroy()
     return true;
 }
 
+bool Backend::clean()
+{
+    bool res = false;
+    // We can have several blocks, occupaing whole region,
+    // because such regions are added in advance (see askMemFromOS() and reset()),
+    // and never used. Release them all.
+    for (int i = advRegBins.getMinUsedBin(0); i != -1; i = advRegBins.getMinUsedBin(i+1)) {
+        if (i == freeAlignedBins.getMinNonemptyBin(i))
+            res |= freeAlignedBins.tryReleaseRegions(i, this);
+        if (i == freeLargeBins.getMinNonemptyBin(i))
+            res |= freeLargeBins.tryReleaseRegions(i, this);
+    }
+
+    scanCoalescQ(/*forceCoalescQDrop=*/false);
+
+    return res;
+}
+
 void Backend::IndexedBins::verify()
 {
     for (int i=0; i<freeBinsNum; i++) {
@@ -1161,14 +1267,18 @@ void Backend::reportStat(FILE *f)
 
     scanCoalescQ(/*forceCoalescQDrop=*/false);
 
+    fprintf(f, "\n  regions:\n");
     {
         MallocMutex::scoped_lock lock(regionListLock);
-        for (MemRegion *curr = regionList; curr; curr = curr->next)
+        for (MemRegion *curr = regionList; curr; curr = curr->next) {
+            fprintf(f, "%p: max block %lu B, ", curr, curr->blockSz);
             regNum++;
+        }
     }
-    fprintf(f, "%d regions\nlarge ", regNum);
+    fprintf(f, "\n%d regions, %lu KB in all regions\n  free bins:\nlarge bins ",
+            regNum, totalMemSize/1024);
     freeLargeBins.reportStat(f);
-    fprintf(f, "\naligned ");
+    fprintf(f, "\naligned bins ");
     freeAlignedBins.reportStat(f);
     fprintf(f, "\n");
 }
