@@ -1,5 +1,5 @@
 /*
-    Copyright 2005-2012 Intel Corporation.  All Rights Reserved.
+    Copyright 2005-2014 Intel Corporation.  All Rights Reserved.
 
     This file is part of Threading Building Blocks.
 
@@ -27,7 +27,6 @@
 */
 
 #include "tbbmalloc_internal.h"
-#include <string.h>
 #include <new>        /* for placement new */
 
 namespace rml {
@@ -45,11 +44,13 @@ struct BackRefBlock : public BlockI {
     // list of all blocks that were allocated from raw mem (i.e., not from backend)
     BackRefBlock *nextRawMemBlock;
     int           allocatedCount; // the number of objects allocated
-    int           myNum;          // the index in the parent array
+    int           myNum;          // the index in the master
     MallocMutex   blockMutex;
-    bool          addedToForUse;  // this block is already added to the listForUse chain
+    // true if this block has been added to the listForUse chain,
+    // modifications protected by masterMutex
+    bool          addedToForUse;
 
-    BackRefBlock(BackRefBlock *blockToUse, int num) :
+    BackRefBlock(const BackRefBlock *blockToUse, int num) :
         nextForUse(NULL), bumpPtr((FreeObject*)((uintptr_t)blockToUse + slabSize - sizeof(void*))),
         freeList(NULL), nextRawMemBlock(NULL), allocatedCount(0), myNum(num),
         addedToForUse(false) {
@@ -57,8 +58,8 @@ struct BackRefBlock : public BlockI {
         // index in BackRefMaster must fit to uint16_t
         MALLOC_ASSERT(!(myNum >> 16), ASSERT_TEXT);
     }
-
-    // TODO: take into account VirtualAlloc granularity
+    // clean all but header
+    void zeroSet() { memset(this+1, 0, BackRefBlock::bytes-sizeof(BackRefBlock)); }
     static const int bytes = slabSize;
 };
 
@@ -76,6 +77,8 @@ struct BackRefMaster {
    taking into account VirtualAlloc allocation granularity */
     static const int leaves = 4;
     static const size_t masterSize = BackRefMaster::bytes+leaves*BackRefBlock::bytes;
+    // take into account VirtualAlloc 64KB granularity
+    static const size_t blockSpaceSize = 64*1024;
 
     Backend       *backend;
     BackRefBlock  *active;         // if defined, use it for allocations
@@ -83,17 +86,19 @@ struct BackRefMaster {
     BackRefBlock  *allRawMemBlocks;
     intptr_t       lastUsed;       // index of the last used block
     bool           rawMemUsed;
+    MallocMutex    requestNewSpaceMutex;
     BackRefBlock  *backRefBl[1];   // the real size of the array is dataSz
 
     BackRefBlock *findFreeBlock();
-    void          addBackRefBlockToList(BackRefBlock *bl);
-    void          addEmptyBackRefBlock(BackRefBlock *newBl);
+    void          addToForUseList(BackRefBlock *bl);
+    void          initEmptyBackRefBlock(BackRefBlock *newBl);
+    bool          requestNewSpace();
 };
 
 const int BackRefMaster::dataSz
     = 1+(BackRefMaster::bytes-sizeof(BackRefMaster))/sizeof(BackRefBlock*);
 
-static MallocMutex backRefMutex;
+static MallocMutex masterMutex;
 static BackRefMaster *backRefMaster;
 
 bool initBackRefMaster(Backend *backend)
@@ -108,11 +113,13 @@ bool initBackRefMaster(Backend *backend)
     master->listForUse = master->allRawMemBlocks = NULL;
     master->rawMemUsed = rawMemUsed;
     master->lastUsed = -1;
+    memset(&master->requestNewSpaceMutex, 0, sizeof(MallocMutex));
     for (int i=0; i<BackRefMaster::leaves; i++) {
-        BackRefBlock *bl = (BackRefBlock *)((uintptr_t)master + BackRefMaster::bytes + i*BackRefBlock::bytes);
-        master->addEmptyBackRefBlock(bl);
+        BackRefBlock *bl = (BackRefBlock*)((uintptr_t)master + BackRefMaster::bytes + i*BackRefBlock::bytes);
+        bl->zeroSet();
+        master->initEmptyBackRefBlock(bl);
         if (i)
-            master->addBackRefBlockToList(bl);
+            master->addToForUseList(bl);
         else // active leaf is not needed in listForUse
             master->active = bl;
     }
@@ -127,7 +134,8 @@ void destroyBackRefMaster(Backend *backend)
         for (BackRefBlock *curr=backRefMaster->allRawMemBlocks; curr; ) {
             BackRefBlock *next = curr->nextRawMemBlock;
             // allRawMemBlocks list is only for raw mem blocks
-            backend->putBackRefSpace(curr, BackRefBlock::bytes, /*rawMemUsed=*/true);
+            backend->putBackRefSpace(curr, BackRefMaster::blockSpaceSize,
+                                     /*rawMemUsed=*/true);
             curr = next;
         }
         backend->putBackRefSpace(backRefMaster, BackRefMaster::masterSize,
@@ -135,18 +143,16 @@ void destroyBackRefMaster(Backend *backend)
     }
 }
 
-void BackRefMaster::addBackRefBlockToList(BackRefBlock *bl)
+void BackRefMaster::addToForUseList(BackRefBlock *bl)
 {
     bl->nextForUse = listForUse;
     listForUse = bl;
     bl->addedToForUse = true;
 }
 
-void BackRefMaster::addEmptyBackRefBlock(BackRefBlock *newBl)
+void BackRefMaster::initEmptyBackRefBlock(BackRefBlock *newBl)
 {
     intptr_t nextLU = lastUsed+1;
-    memset((char*)newBl+sizeof(BackRefBlock), 0,
-           BackRefBlock::bytes-sizeof(BackRefBlock));
     new (newBl) BackRefBlock(newBl, nextLU);
     backRefBl[nextLU] = newBl;
     // lastUsed is read in getBackRef, and access to backRefBl[lastUsed]
@@ -154,29 +160,61 @@ void BackRefMaster::addEmptyBackRefBlock(BackRefBlock *newBl)
     FencedStore(lastUsed, nextLU);
 }
 
+bool BackRefMaster::requestNewSpace()
+{
+    bool rawMemUsed;
+    MALLOC_STATIC_ASSERT(!(blockSpaceSize % BackRefBlock::bytes),
+                         "Must request space for whole number of blocks.");
+
+    // only one thread at a time may add blocks
+    MallocMutex::scoped_lock newSpaceLock(requestNewSpaceMutex);
+
+    if (listForUse) // double check that only one block is available
+        return true;
+    BackRefBlock *newBl =
+        (BackRefBlock*)backend->getBackRefSpace(blockSpaceSize, &rawMemUsed);
+    if (!newBl) return false;
+
+    // touch a page for the 1st time without taking masterMutex ...
+    for (BackRefBlock *bl = newBl; (uintptr_t)bl < (uintptr_t)newBl + blockSpaceSize;
+         bl = (BackRefBlock*)((uintptr_t)bl + BackRefBlock::bytes))
+        bl->zeroSet();
+
+    MallocMutex::scoped_lock lock(masterMutex); // ... and share under lock
+    // use the first block in the batch to maintain the list of "raw" memory
+    // to be released at shutdown
+    if (rawMemUsed) {
+        newBl->nextRawMemBlock = backRefMaster->allRawMemBlocks;
+        backRefMaster->allRawMemBlocks = newBl;
+    }
+    for (BackRefBlock *bl = newBl; (uintptr_t)bl < (uintptr_t)newBl + blockSpaceSize;
+         bl = (BackRefBlock*)((uintptr_t)bl + BackRefBlock::bytes)) {
+        initEmptyBackRefBlock(bl);
+        if (active->allocatedCount == BR_MAX_CNT)
+            active = bl; // active leaf is not needed in listForUse
+        else
+            addToForUseList(bl);
+    }
+    return true;
+}
+
 BackRefBlock *BackRefMaster::findFreeBlock()
 {
     if (active->allocatedCount < BR_MAX_CNT)
         return active;
 
     if (listForUse) {                                   // use released list
-        active = listForUse;
-        listForUse = listForUse->nextForUse;
-        MALLOC_ASSERT(active->addedToForUse, ASSERT_TEXT);
-        active->addedToForUse = false;
+        MallocMutex::scoped_lock lock(masterMutex);
+
+        if (active->allocatedCount == BR_MAX_CNT && listForUse) {
+            active = listForUse;
+            listForUse = listForUse->nextForUse;
+            MALLOC_ASSERT(active->addedToForUse, ASSERT_TEXT);
+            active->addedToForUse = false;
+        }
     } else if (lastUsed-1 < backRefMaster->dataSz) {    // allocate new data node
-        bool rawMemUsed;
-        BackRefBlock *newBl =
-            (BackRefBlock*)backend->getBackRefSpace(BackRefBlock::bytes, &rawMemUsed);
-        if (!newBl) return NULL;
-        backRefMaster->addEmptyBackRefBlock(newBl);
-        if (rawMemUsed) {
-            newBl->nextRawMemBlock = backRefMaster->allRawMemBlocks;
-            backRefMaster->allRawMemBlocks = newBl;
-        } else
-            newBl->nextRawMemBlock = NULL;
-        active = newBl;
-    } else  // no free blocks, give up
+        if (!requestNewSpace()) return NULL;
+    } else // no free space in BackRefMaster, give up
         return NULL;
     return active;
 }
@@ -206,16 +244,13 @@ BackRefIdx BackRefIdx::newBackRef(bool largeObj)
     BackRefBlock *blockToUse;
     void **toUse;
     BackRefIdx res;
+    bool lastBlockFirstUsed = false;
 
     do {
-        { // global lock taken to find a block
-            MallocMutex::scoped_lock lock(backRefMutex);
-
-            MALLOC_ASSERT(backRefMaster, ASSERT_TEXT);
-            blockToUse = backRefMaster->findFreeBlock();
-            if (!blockToUse)
-                return BackRefIdx();
-        }
+        MALLOC_ASSERT(backRefMaster, ASSERT_TEXT);
+        blockToUse = backRefMaster->findFreeBlock();
+        if (!blockToUse)
+            return BackRefIdx();
         toUse = NULL;
         { // the block is locked to find a reference
             MallocMutex::scoped_lock lock(blockToUse->blockMutex);
@@ -238,10 +273,18 @@ BackRefIdx BackRefIdx::newBackRef(bool largeObj)
                     blockToUse->bumpPtr = NULL;
                 }
             }
-            if (toUse)
+            if (toUse) {
+                if (!blockToUse->allocatedCount && !backRefMaster->listForUse)
+                    lastBlockFirstUsed = true;
                 blockToUse->allocatedCount++;
+            }
         } // end of lock scope
     } while (!toUse);
+    // The first thread that uses the last block requests new space in advance;
+    // possible failures are ignored.
+    if (lastBlockFirstUsed)
+        backRefMaster->requestNewSpace();
+
     res.master = blockToUse->myNum;
     uintptr_t offset =
         ((uintptr_t)toUse - ((uintptr_t)blockToUse + sizeof(BackRefBlock)))/sizeof(void*);
@@ -276,10 +319,10 @@ void removeBackRef(BackRefIdx backRefIdx)
     }
     // TODO: do we need double-check here?
     if (!currBlock->addedToForUse && currBlock!=backRefMaster->active) {
-        MallocMutex::scoped_lock lock(backRefMutex);
+        MallocMutex::scoped_lock lock(masterMutex);
 
         if (!currBlock->addedToForUse && currBlock!=backRefMaster->active)
-            backRefMaster->addBackRefBlockToList(currBlock);
+            backRefMaster->addToForUseList(currBlock);
     }
 }
 
