@@ -1,5 +1,5 @@
 /*
-    Copyright 2005-2012 Intel Corporation.  All Rights Reserved.
+    Copyright 2005-2014 Intel Corporation.  All Rights Reserved.
 
     This file is part of Threading Building Blocks.
 
@@ -642,60 +642,40 @@ void TestConcurrentGrowBy( int nthread ) {
 //TODO: move this to more appropriate place, smth like test_harness.cpp
 void TestArrayLength(){
     int five_elementh_array[5] = {0};
-    ASSERT(array_length(five_elementh_array)==5,"array_length failed to determine length of non empty non dynamic array");
+    ASSERT(Harness::array_length(five_elementh_array)==5,"array_length failed to determine length of non empty non dynamic array");
 }
 
 #if __TBB_INITIALIZER_LISTS_PRESENT
-//TODO: move init list test set to separate header
-//TODO: split into set of tests
-//TODO: add test for no leaks, and correct element lifetime
-//the need for macro comes from desire to test different scenarios where initializer sequence is compile time constant
-#define __TBB_TEST_INIT_LIST_SUITE(FUNC_NAME, CONTAINER, ELEMENT_TYPE, INIT_SEQ)                                                                  \
-void FUNC_NAME(){                                                                                                                                 \
-    typedef ELEMENT_TYPE element_type;                                                                                                            \
-    typedef CONTAINER<element_type> container_type;                                                                                               \
-    element_type test_seq[] = INIT_SEQ;                                                                                                           \
-    container_type expected(test_seq,test_seq + array_length(test_seq));                                                                          \
-                                                                                                                                                  \
-    /*test for explicit contructor call*/                                                                                                         \
-    container_type vd (INIT_SEQ,tbb::cache_aligned_allocator<int>());                                                                             \
-    ASSERT(vd == expected,"initialization via explicit constructor call with init list failed");                                                  \
-    /*test for explicit contructor call with std::initializer_list*/                                                                              \
-                                                                                                                                                  \
-    std::initializer_list<element_type> init_list = INIT_SEQ;                                                                                     \
-    container_type v1 (init_list,tbb::cache_aligned_allocator<int>());                                                                            \
-    ASSERT(v1 == expected,"initialization via explicit constructor call with std::initializer_list failed");                                      \
-                                                                                                                                                  \
-    /*implicit constructor call test*/                                                                                                            \
-    container_type v = INIT_SEQ;                                                                                                                  \
-    ASSERT(v == expected,"init list constructor failed");                                                                                         \
-                                                                                                                                                  \
-    /*assignment operator test*/                                                                                                                  \
-    /*TODO: count created and destroyed injects to assert that no extra copy of vector was created implicitly*/                                   \
-    container_type va;                                                                                                                            \
-    va = INIT_SEQ;                                                                                                                                \
-    ASSERT(va == expected,"init list operator= failed");                                                                                          \
-                                                                                                                                                  \
-    container_type vae;                                                                                                                           \
-    vae.assign(INIT_SEQ);                                                                                                                          \
-    ASSERT(vae == expected,"init list assign failed");                                                                                             \
-}
-
-#define __TBB_CVECTOR_TEST_INIT_SEQ {1,2,3,4,5}
-__TBB_TEST_INIT_LIST_SUITE(TestInitListIml,tbb::concurrent_vector,char,__TBB_CVECTOR_TEST_INIT_SEQ )
-#undef __TBB_CVECTOR_TEST_INIT_SEQ
-
-#define __TBB_CVECTOR_TEST_EMPTY_INIT_SEQ {}
-__TBB_TEST_INIT_LIST_SUITE(TestEmptyInitListIml,tbb::concurrent_vector,int,__TBB_CVECTOR_TEST_EMPTY_INIT_SEQ )
-#undef __TBB_CVECTOR_TEST_EMPTY_INIT_SEQ
+#include "test_initializer_list.h"
 
 void TestInitList(){
     REMARK("testing initializer_list methods \n");
-    TestEmptyInitListIml();
-    TestInitListIml();
+    using namespace initializer_list_support_tests;
+    TestInitListSupport<tbb::concurrent_vector<char> >({1,2,3,4,5});
+    TestInitListSupport<tbb::concurrent_vector<int> >({});
 }
 #endif //if __TBB_INITIALIZER_LISTS_PRESENT
 
+#if __TBB_RANGE_BASED_FOR_PRESENT
+#include "test_range_based_for.h"
+#include <functional>
+
+void TestRangeBasedFor(){
+    using namespace range_based_for_support_tests;
+
+    REMARK("testing range based for loop compatibility \n");
+    typedef tbb::concurrent_vector<int> c_vector;
+    c_vector a_c_vector;
+
+    const int sequence_length = 100;
+    for (int i =1; i<= sequence_length; ++i){
+        a_c_vector.push_back(i);
+    }
+
+    ASSERT( range_based_for_accumulate(a_c_vector, std::plus<int>(), 0) == gauss_summ_of_int_sequence(sequence_length), "incorrect accumulated value generated via range based for ?");
+}
+#endif //if __TBB_RANGE_BASED_FOR_PRESENT
+
 //! Test the assignment operator and swap
 void TestAssign() {
     typedef tbb::concurrent_vector<FooWithAssign, local_counting_allocator<std::allocator<FooWithAssign>, size_t > > vector_t;
@@ -1016,11 +996,11 @@ void TestExceptions() {
 #endif /* TBB_USE_EXCEPTIONS */
 
 //------------------------------------------------------------------------
-// Test SSE / AVX
+// Test support for SIMD instructions
 //------------------------------------------------------------------------
 #include "harness_m128.h"
 
-#if HAVE_m128 | HAVE_m256
+#if HAVE_m128 || HAVE_m256
 
 template<typename ClassWithVectorType>
 void TestVectorTypes() {
@@ -1073,6 +1053,9 @@ int TestMain () {
     TestComparison();
     TestFindPrimes();
     TestSort();
+#if __TBB_RANGE_BASED_FOR_PRESENT
+    TestRangeBasedFor();
+#endif //if __TBB_RANGE_BASED_FOR_PRESENT
 #if __TBB_THROW_ACROSS_MODULE_BOUNDARY_BROKEN
     REPORT("Known issue: exception safety test is skipped.\n");
 #elif TBB_USE_EXCEPTIONS
