@@ -1,5 +1,5 @@
 /*
-    Copyright 2005-2012 Intel Corporation.  All Rights Reserved.
+    Copyright 2005-2014 Intel Corporation.  All Rights Reserved.
 
     This file is part of Threading Building Blocks.
 
@@ -29,9 +29,6 @@
 #include "harness_graph.h"
 
 #include "tbb/task_scheduler_init.h"
-#include "tbb/spin_mutex.h"
-
-tbb::spin_mutex global_mutex;
 
 #define N 1000
 #define MAX_NODES 4
@@ -43,6 +40,14 @@ struct empty_no_assign : private NoAssign {
    operator int() { return 0; }
 };
 
+// A class to use as a fake predecessor of continue_node
+struct fake_continue_sender : public tbb::flow::sender<tbb::flow::continue_msg>
+{
+    // Define implementations of virtual methods that are abstract in the base class
+    /*override*/ bool register_successor( successor_type& ) { return false; }
+    /*override*/ bool remove_successor( successor_type& )   { return false; }  
+};
+
 template< typename InputType >
 struct parallel_puts : private NoAssign {
 
@@ -61,8 +66,9 @@ struct parallel_puts : private NoAssign {
 
 template< typename OutputType >
 void run_continue_nodes( int p, tbb::flow::graph& g, tbb::flow::continue_node< OutputType >& n ) {
+    fake_continue_sender fake_sender;
     for (size_t i = 0; i < N; ++i) {
-        n.register_predecessor( *reinterpret_cast< tbb::flow::sender< tbb::flow::continue_msg > * >(&n) );
+        n.register_predecessor( fake_sender );
     }
 
     for (size_t num_receivers = 1; num_receivers <= MAX_NODES; ++num_receivers ) {
@@ -127,8 +133,9 @@ void continue_nodes_with_copy( ) {
         global_execute_count = Offset;
 
         tbb::flow::continue_node< OutputType > exe_node( g, cf );
+        fake_continue_sender fake_sender;
         for (size_t i = 0; i < N; ++i) {
-           exe_node.register_predecessor( *reinterpret_cast< tbb::flow::sender< tbb::flow::continue_msg > * >(&exe_node) );
+           exe_node.register_predecessor( fake_sender );
         }
 
         for (size_t num_receivers = 1; num_receivers <= MAX_NODES; ++num_receivers ) {
@@ -177,6 +184,55 @@ void test_concurrency(int num_threads) {
     run_continue_nodes<int>();
     run_continue_nodes<empty_no_assign>();
 }
+/*
+ * Connection of two graphs is not currently supported, but works to some limited extent.
+ * This test is included to check for backward compatibility. It checks that a continue_node 
+ * with predecessors in two different graphs receives the required
+ * number of continue messages before it executes.
+ */
+using namespace tbb::flow;
+
+struct add_to_counter {
+    int* counter;
+    add_to_counter(int& var):counter(&var){}
+    void operator()(continue_msg){*counter+=1;}
+};
+
+void test_two_graphs(){
+    int count=0;
+
+    //graph g with broadcast_node and continue_node
+    graph g;
+    broadcast_node<continue_msg> start_g(g);
+    continue_node<continue_msg> first_g(g, add_to_counter(count));
+
+    //graph h with broadcast_node
+    graph h;
+    broadcast_node<continue_msg> start_h(h);
+
+    //making two edges to first_g from the two graphs
+    make_edge(start_g,first_g);
+    make_edge(start_h, first_g);
+
+    //two try_puts from the two graphs
+    start_g.try_put(continue_msg());
+    start_h.try_put(continue_msg());
+    g.wait_for_all();
+    ASSERT(count==1, "Not all continue messages received");
+ 
+    //two try_puts from the graph that doesn't contain the node
+    count=0;
+    start_h.try_put(continue_msg());
+    start_h.try_put(continue_msg());
+    g.wait_for_all();
+    ASSERT(count==1, "Not all continue messages received -1");
+
+    //only one try_put
+    count=0;
+    start_g.try_put(continue_msg());
+    g.wait_for_all();
+    ASSERT(count==0, "Node executed without waiting for all predecessors");
+}
 
 int TestMain() {
     if( MinThread<1 ) {
@@ -186,6 +242,7 @@ int TestMain() {
     for( int p=MinThread; p<=MaxThread; ++p ) {
        test_concurrency(p);
    }
+   test_two_graphs();
    return Harness::Done;
 }
 
