@@ -1,5 +1,5 @@
 /*
-    Copyright 2005-2012 Intel Corporation.  All Rights Reserved.
+    Copyright 2005-2014 Intel Corporation.  All Rights Reserved.
 
     This file is part of Threading Building Blocks.
 
@@ -29,6 +29,20 @@
 const unsigned MByte = 1024*1024;
 bool __tbb_test_errno = false;
 
+#include "tbb/tbb_config.h"
+
+#if __TBB_WIN8UI_SUPPORT	
+// testing allocator itself not iterfaces
+// so we can use desktop functions
+#define _CRT_USE_WINAPI_FAMILY_DESKTOP_APP 1
+#define HARNESS_NO_PARSE_COMMAND_LINE 1
+#include "harness.h"
+// FIXME: fix the test to support New Windows *8 Store Apps mode.
+int TestMain() {
+    return Harness::Skipped;
+}
+#else /* __TBB_WIN8UI_SUPPORT	 */
+
 /* _WIN32_WINNT should be defined at the very beginning,
    because other headers might include <windows.h>
 */
@@ -98,9 +112,12 @@ void limitMem( size_t limit )
 #include "tbb/scalable_allocator.h"
 
 #define HARNESS_CUSTOM_MAIN 1
+#define HARNESS_TBBMALLOC_THREAD_SHUTDOWN 1
 #include "harness.h"
 #include "harness_barrier.h"
+#if !__TBB_SOURCE_DIRECTLY_INCLUDED
 #include "harness_tbb_independence.h"
+#endif
 #if __linux__
 #include <stdint.h> // uintptr_t
 #endif
@@ -167,7 +184,7 @@ TestAlignedRealloc* Raligned_realloc;
 bool error_occurred = false;
 
 #if __APPLE__
-// Tests that use the variable are skipped on Mac OS* X
+// Tests that use the variable are skipped on OS X*
 #else
 static bool perProcessLimits = true;
 #endif
@@ -248,7 +265,7 @@ static void setSystemAllocs()
     Taligned_free=_aligned_free;
     Rposix_memalign=0;
 #elif  __APPLE__ || __sun || __ANDROID__ 
-// Mac OS* X, Solaris, and Android don't have posix_memalign
+// OS X*, Solaris, and Android don't have posix_memalign
     Raligned_malloc=0;
     Raligned_realloc=0;
     Taligned_free=0;
@@ -308,9 +325,17 @@ int main(int argc, char* argv[]) {
     for (int i=1; i< argc; i++) {
         if (strcmp((char*)*(argv+i),"-s")==0)
         {
+#if __INTEL_COMPILER == 1400 && __linux__
+            // Workaround for Intel(R) C++ Compiler XE, version 14.0.0.080:
+            // unable to call setSystemAllocs() in such configuration.
+            REPORT("Known issue: Standard allocator testing is not supported.\n");
+            REPORT( "skip\n" );
+            return 0;
+#else
             setSystemAllocs();
             argC--;
             break;
+#endif
         }
     }
 
@@ -325,7 +350,7 @@ int main(int argc, char* argv[]) {
 #endif
     //-------------------------------------
 #if __APPLE__
-    /* Skip due to lack of memory limit enforcing under Mac OS X. */
+    /* Skip due to lack of memory limit enforcing under OS X*. */
 #else
     limitMem(200);
     ReallocParam();
@@ -362,10 +387,17 @@ int main(int argc, char* argv[]) {
 
     for( int p=MaxThread; p>=MinThread; --p ) {
         REMARK("testing with %d threads\n", p );
-        Harness::SpinBarrier *barrier = new Harness::SpinBarrier(p);
-        NativeParallelFor( p, RoundRobin(p, barrier, Verbose) );
-        delete barrier;
+        for (int limit=0; limit<2; limit++) {
+            int ret = scalable_allocation_mode(TBBMALLOC_SET_SOFT_HEAP_LIMIT,
+                                               16*1024*limit);
+            ASSERT(ret==TBBMALLOC_OK, NULL);
+            Harness::SpinBarrier *barrier = new Harness::SpinBarrier(p);
+            NativeParallelFor( p, RoundRobin(p, barrier, Verbose) );
+            delete barrier;
+        }
     }
+    int ret = scalable_allocation_mode(TBBMALLOC_SET_SOFT_HEAP_LIMIT, 0);
+    ASSERT(ret==TBBMALLOC_OK, NULL);
     if( !error_occurred )
         REPORT("done\n");
     return 0;
@@ -1003,7 +1035,7 @@ void CMemTest::RunAllTests(int total_threads)
         InvariantDataRealloc(/*aligned=*/true);
     TestAlignedParameters();
 #if __APPLE__
-    REPORT("Known issue: some tests are skipped on Mac OS* X\n");
+    REPORT("Known issue: some tests are skipped on OS X*\n");
 #else
     UniquePointer();
     AddrArifm();
@@ -1013,3 +1045,5 @@ void CMemTest::RunAllTests(int total_threads)
 #endif
     if (FullLog) REPORT("Tests for %d threads ended\n", total_threads);
 }
+
+#endif /* __TBB_WIN8UI_SUPPORT	 */
